import typer

from chroma_ops.wal_commit import command as commit_wal_command
from chroma_ops.wal_clean import command as clean_wal_command
from chroma_ops.wal_export import command as export_wal_command

app = typer.Typer(no_args_is_help=True, help="ChromaDB Ops Commands.")

app.command(name="commit-wal", help="Commit WAL to HNSW lib binary index", no_args_is_help=True)(
    commit_wal_command
)

app.command(name="clean-wal", help="Cleans up WAL and VACUUM the SQLite DB.", no_args_is_help=True)(
    clean_wal_command
)
app.command(name="export-wal", help="Exports the WAL to a jsonl file.", no_args_is_help=True)(
    export_wal_command
)
if __name__ == "__main__":
    app()



import argparse
import base64
import json
import os
import sqlite3
import sys
from contextlib import contextmanager

import typer

from chroma_ops.utils import validate_chroma_persist_dir


@contextmanager
def smart_open(filename=None):
    if filename:
        fh = open(filename, 'w')
    else:
        fh = sys.stdout

    try:
        yield fh
    finally:
        if filename:
            fh.close()


def export_wal(persist_dir: str, output_file: str):
    validate_chroma_persist_dir(persist_dir)
    sql_file = os.path.join(persist_dir, "chroma.sqlite3")
    conn = sqlite3.connect(f"file:{sql_file}?mode=ro", uri=True)
    cursor = conn.cursor()
    query = "SELECT * FROM embeddings_queue ORDER BY seq_id ASC;"
    cursor.execute(query)
    exported_rows = 0
    with smart_open(output_file) as json_file:
        column_names = [description[0] for description in cursor.description]
        for row in cursor:
            row_data = {}
            for i, column_name in enumerate(column_names):
                if column_name == 'vector':
                    row_data[column_name] = base64.b64encode(row[i]).decode()
                else:
                    row_data[column_name] = row[i]
            json_file.write(json.dumps(row_data) + "\n")
            exported_rows += 1

    conn.close()
    typer.echo(f"Exported {exported_rows} rows", file=sys.stderr)


def command(
        persist_dir: str = typer.Argument(..., help="The persist directory"),
        out: str = typer.Option(None, "--out", "-o", help="The output jsonl file"),
):
    export_wal(persist_dir, out)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("persist_dir", type=str)
    parser.add_argument("--out", type=str, default=None)
    arg = parser.parse_args()
    export_wal(arg.persist_dir, arg.out)



#!/usr/bin/env python3
import argparse
import os
import sqlite3
import hnswlib
import typer
from chromadb.segment.impl.vector.local_persistent_hnsw import PersistentData

from chroma_ops.utils import validate_chroma_persist_dir, get_hnsw_index_ids, get_dir_size


def clean_wal(persist_dir: str):
    validate_chroma_persist_dir(persist_dir)
    print("Size before: ", get_dir_size(persist_dir))
    # TODO add path join here
    sql_file = os.path.join(persist_dir, "chroma.sqlite3")
    conn = sqlite3.connect(sql_file)
    # conn = sqlite3.connect(f"file:{sql_file}?mode=ro", uri=True)

    cursor = conn.cursor()

    query = "SELECT s.id as 'segment',s.topic as 'topic', c.id as 'collection' , c.dimension as 'dimension' FROM segments s LEFT JOIN collections c ON s.collection = c.id WHERE s.scope = 'VECTOR';"

    cursor.execute(query)

    results = cursor.fetchall()
    wal_cleanup_queries = []
    for row in results:
        metadata_pickle = os.path.join(persist_dir, row[0], 'index_metadata.pickle')
        if os.path.exists(metadata_pickle):
            metadata = PersistentData.load_from_file(metadata_pickle)
            wal_cleanup_queries.append(
                f"DELETE FROM embeddings_queue WHERE seq_id < {metadata.max_seq_id} AND topic='{row[1]}';"
            )
        else:
            hnsw_space = cursor.execute(
                "select str_value from collection_metadata where collection_id=? and key='hnsw:space'",
                (row[2],),
            ).fetchone()
            hnsw_space = "l2" if hnsw_space is None else hnsw_space[0]
            list_of_ids = get_hnsw_index_ids(
                f"{os.path.join(persist_dir, row[0])}", hnsw_space, row[3]
            )
            batch_size = 100
            for batch in range(0, len(list_of_ids), batch_size):
                wal_cleanup_queries.append(
                    f"DELETE FROM embeddings_queue WHERE seq_id IN ({','.join([str(i) for i in list_of_ids[batch:batch + batch_size]])});"
                )
    if len(wal_cleanup_queries) > 0:
        print("Cleaning up WAL")
        wal_cleanup_queries.append("VACUUM;")
        cursor.executescript("\n".join(wal_cleanup_queries))
    # Close the cursor and connection
    cursor.close()
    conn.close()
    print("Size after: ", get_dir_size(persist_dir))


def command(
        persist_dir: str = typer.Argument(..., help="The persist directory"),
):
    clean_wal(persist_dir)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("persist_dir", type=str)
    arg = parser.parse_args()
    clean_wal(arg.persist_dir)



#!/usr/bin/env python3
import argparse
import os

import chromadb
import typer

from chromadb.segment.impl.vector.local_persistent_hnsw import PersistentData
from chromadb.segment import VectorReader
from chromadb.types import SegmentScope, Operation

from chroma_ops.utils import validate_chroma_persist_dir


def commit_wal(persist_dir: str):
    """Note this uses internal ChromaDB APIs which may change at any moment."""

    validate_chroma_persist_dir(persist_dir)
    client = chromadb.PersistentClient(path=persist_dir)
    vector_segments = [
        s
        for s in client._server._sysdb.get_segments()
        if s["scope"] == SegmentScope.VECTOR
    ]
    for s in vector_segments:
        col = client._server._get_collection(
            s["collection"]
        )  # load the collection and apply WAL
        client._server._manager.hint_use_collection(
            s["collection"], Operation.ADD
        )  # Add hint to load the index into memory
        segment = client._server._manager.get_segment(
            s["collection"], VectorReader
        )  # Get the segment instance
        segment._apply_batch(segment._curr_batch)  # Apply the current WAL batch

        segment._persist()  # persist the index
        metadata = PersistentData.load_from_file(
            os.path.join(persist_dir, str(s["id"]), "index_metadata.pickle")
        )  # load the metadata after persisting
        print(
            f"Processing index for collection {col['name']} ({s['id']}) - "
            f"total vectors in index {len(segment._index.get_ids_list())}"
        )
        segment.close_persistent_index()


def command(
        persist_dir: str = typer.Argument(..., help="The persist directory"),
):
    commit_wal(persist_dir)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("persist_dir", type=str)
    arg = parser.parse_args()
    commit_wal(arg.persist_dir)



import os
from typing import List

import hnswlib


def validate_chroma_persist_dir(persist_dir: str) -> None:
    if not os.path.exists(persist_dir):
        raise ValueError(f"Persist dir ({persist_dir}) does not exist")
    if not os.path.exists(f"{persist_dir}/chroma.sqlite3"):
        raise ValueError(
            f"{persist_dir} does not appear to be valid ChromaDB persist directory"
        )


def get_hnsw_index_ids(filename: str, space: str = "l2", dim: int = 384) -> List[int]:
    index = hnswlib.Index(space=space, dim=dim)
    index.load_index(
        filename,
        is_persistent_index=True,
        max_elements=100000,
    )
    ids = index.get_ids_list().copy()
    index.close_file_handles()
    return ids


def get_dir_size(path: str) -> int:
    total_size = 0
    for dirpath, dirnames, filenames in os.walk(path):
        for f in filenames:
            fp = os.path.join(dirpath, f)
            if os.path.exists(fp):
                total_size += os.path.getsize(fp)
    return total_size
